/*
** Module   : InternalEntry.java
** Abstract : stores metadata relating to an internal entry 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 GES 20131124 Moved this code out of SourceNameMapper into its own class. This enables
**                  cleaner separation and modularity. Added synchronization and made changes
**                  for RETURN support.
*/

package com.goldencode.p2j.util;

import java.util.*;

/**
 * Container for legacy info belonging to an internal entry (a user-defined internal procedure,
 * function or external API call).
 */
public class InternalEntry
{
   /** Identify a native procedure. */
   public static final String TYPE_DLL_ENTRY = "DLL-ENTRY";

   /** Identify an internal procedure. */
   public static final String TYPE_PROCEDURE = "PROCEDURE";

   /** Identify an user defined function. */
   public static final String TYPE_FUNCTION  = "FUNCTION";
   
   /** Identify parameters in INPUT mode. */
   public static final char INPUT_MODE = 'I';

   /** Identify parameters in INPUT-OUTPUT mode. */
   public static final char INPUT_OUTPUT_MODE = 'U';

   /** Identify parameters in OUTPUT mode. */
   public static final char OUTPUT_MODE = 'O';

   /** Identify buffer parameters. */
   public static final char BUFFER_MODE = 'B';

   /** Legacy 4GL name for this internal-entry. */
   final String pname;

   /** Converted Java name for this internal-entry. */
   final String jname;
   
   /** The type of this internal entry. */
   private final String type;
   
   /**
    * The signature of this internal entry, as returned by the 
    * <code>get-signature</code> 4GL API.
    */
   private String signature = null;
   
   /**
    * The string representation of the parameter modes for this internal
    * entry. 
    */
   private String parameterModes = null;
   
   /** The parameter list of this internal entry. */
   private List<Parameter> parameters = Collections.EMPTY_LIST;

   /** A map containing all the internal entry's attributes. */
   private Map<String, String> attributes = new HashMap<String, String>();
   
   /** 
    * Flag indicating if this is a function (when set to 
    * <code>true</code>). Gets populated on first {@link #isFunction()} call.
    */
   private Boolean function = null;

   /** 
    * Flag indicating if this is a IN SUPER internal entry (when set to 
    * <code>true</code>). Gets populated on first {@link #isSuper()} call.
    */
   private Boolean isSuper = null;

   /** 
    * Flag indicating if this is a IN handle internal entry (when set to 
    * <code>true</code>). Gets populated on first {@link #isInHandle()()} 
    * call.
    */
   private Boolean inHandle = null;

   /** 
    * Flag indicating if this is a PRIVATE internal entry (when set to 
    * <code>true</code>). Gets populated on first {@link #isPrivate()()} 
    * call.
    */
   private Boolean isPrivate = null;

   /**
    * Basic c'tor.
    * 
    * @param    pname
    *           Legacy 4GL name for this internal-entry.
    * @param    jname
    *           Converted Java name for this internal-entry.
    * @param    type
    *           The type of this internal entry.
    */
   public InternalEntry(String pname, String jname, String type)
   {
      this.pname = pname;
      this.jname = jname;
      this.type  = type;
   }
   
   /**
    * Check if the given mode is an OUTPUT mode (i.e. {@link #OUTPUT_MODE} or 
    * {@link #INPUT_OUTPUT_MODE}.
    * 
    * @return   <code>true</code> if the given character represents one of the output
    *           modes.
    */
   public static boolean isOutputMode(char mode)
   {
      return mode == OUTPUT_MODE || mode == INPUT_OUTPUT_MODE;
   }
   
   /**
    * Return <code>true</code> if this internal-entry has its 
    * <code>in-super</code> attribute set to <code>"true"</code>.
    * 
    * @return   See above.
    */
   public synchronized boolean isSuper()
   {
      if (isSuper == null)
      {
         isSuper = "true".equalsIgnoreCase(getAttribute("in-super"));
      }
      
      return isSuper;
   }
   
   /**
    * Return <code>true</code> if this internal-entry has its 
    * <code>private</code> attribute set to <code>"true"</code>.
    * 
    * @return   See above.
    */
   public synchronized boolean isPrivate()
   {
      if (isPrivate == null)
      {
         isPrivate = "true".equalsIgnoreCase(getAttribute("private"));
      }
      
      return isPrivate;
   }
   
   /**
    * Return <code>true</code> if internal-entry is a function (its 
    * {@link #type} is set to {@link #TYPE_FUNCTION}).
    * 
    * @return   See above.
    */
   public synchronized boolean isFunction()
   {
      if (function == null)
      {
         function = TYPE_FUNCTION.equalsIgnoreCase(type);
      }
      
      return function;
   }

   /**
    * Return <code>true</code> if this internal-entry has its 
    * <code>in-handle</code> attribute set to <code>"true"</code>.
    * 
    * @return   See above.
    */
   public synchronized boolean isInHandle()
   {
      if (inHandle == null)
      {
         inHandle = "true".equalsIgnoreCase(getAttribute("in-handle"));
      }

      return inHandle;
   }
   
   /**
    * Obtain the original name of the procedure being executed as specified in the 4GL
    * code.
    *
    * @return   The original (unconverted) name.
    */
   public String getLegacyName()
   {
      return pname;
   }

   /**
    * Get the value of the given attribute.  If the attribute is null or
    * is not set, return <code>null</code>.
    * 
    * @param    attr
    *           The attribute's name.
    *           
    * @return   See above.
    */
   public synchronized String getAttribute(String attr)
   {
      if (attr == null)
      {
         return null;
      }
      
      attr = attr.toLowerCase();
      
      return attributes.get(attr);
   }

   /**
    * Set the value of the given attribute.
    * 
    * @param    name
    *           The attribute's name.
    * @param    value
    *           The attribute's value.
    */
   public synchronized void putAttribute(String name, String value)
   {
      attributes.put(name.toLowerCase(), value);
   }

   /**
    * Gets the signature of this internal-entry, as returned by the <code>get-signature</code>
    * API of an persistent procedure. If this is the first access, it initializes the 
    * {@link #signature} field.
    * 
    * @return   The 4GL-compatible signature string.
    */
   public synchronized String getSignature()
   {
      if (signature == null)
      {
         signature = buildSignature();
      }
      
      return signature;
   }

   /**
    * Gets the details of the given parameter.
    *
    * @param    idx
    *           0-based parameter index.
    * 
    * @return   The parameter instance or <code>null</code> if no such parameter exists.
    */
   public synchronized Parameter getParameter(int idx)
   {
      if (idx < 0 || idx > parameters.size())
      {
         return null;
      }
      
      return parameters.get(idx);
   }
   
   /**
    * Returns the number of parameters for this internal entry.
    *
    * @return   The size of the parameter list.
    */
   public synchronized int getParameterListSize()
   {
      return parameters.size();
   }
   
   /**
    * Returns the list of parameters for this internal entry.
    *
    * @return   The parameter list.
    */
   public synchronized List<Parameter> getParameterList()
   {
      // this is safe because the list cannot be modified
      return parameters;
   }
   
   /**
    * Gets the string representation of the parameter modes for 
    * internal-entry. If this is the first access, it initializes the 
    * {@link #parameterModes} field.
    * 
    * @return   See above.
    */
   public synchronized String getParameterModes()
   {
      if (parameterModes == null)
      {
         if (parameters.size() == 0)
         {
            return null;
         }
         else
         {
            StringBuilder sb = new StringBuilder();
            
            for (Parameter p : parameters)
            {
               String pmode = p.getMode();
               String type = p.getType();
               char mode = 0;

               if ("BUFFER".equalsIgnoreCase(type))
               {
                  mode = BUFFER_MODE;
               }
               else if ("INPUT".equalsIgnoreCase(pmode))
               {
                  mode = INPUT_MODE;
               }
               else if ("INPUT-OUTPUT".equalsIgnoreCase(pmode))
               {
                  mode = INPUT_OUTPUT_MODE;
               }
               else if ("OUTPUT".equalsIgnoreCase(pmode) || "RETURN".equalsIgnoreCase(pmode))
               {
                  // the resulting string is used for validation purposes, so we must convert
                  // RETURN parms to OUTPUT (although since we defer validation to 
                  // NativeInvoker.invoke(), I'm not sure this is actually needed)
                  mode = OUTPUT_MODE;
               }
               else
               {
                  String err = "Malformed parameter mode " + pmode + " for parameter " + pname;
                  throw new IllegalStateException(err);
               }
               
               sb.append(mode);
            }
            
            parameterModes = sb.toString();
         }
      }
      
      return parameterModes;
   }
   
   /**
    * Sets the list of parameters for this internal entry.  Changes to the input list will
    * not be reflected in the saved list.
    *
    * @param    parameters
    *           The parameter list.
    */
   synchronized void setParameterList(List<Parameter> parameters)
   {
      this.parameters = Collections.unmodifiableList(parameters);
   }
   
   /**
    * Build the 4GL-style signature of this instance.
    * 
    * @return   The internal-entry's signature.
    */
   private String buildSignature()
   {
      // The format is: TYPE,return type,param1,param2,param3 where:
      // - TYPE is either PROCEDURE, FUNCTION, EXTERN or DLL-ENTRY
      // - return type is the 4GL name of the return data type
      // - param# is one of: 
      //     INPUT varname TYPE
      //     OUTPUT varname TYPE or INPUT-OUTPUT varname TYPE.
      //   Varname is either the original 4GL variable name or "arg#", if this
      //   is a header definition (i.e. an IN SUPER or IN handle) which had
      //   its parameter set as:
      //   "function f1 returns int (input int) in super."
      
      String type    = this.type;
      String returns = getAttribute("returns");
      
      if (returns == null)
      {
         returns = "";
      }
      
      if (isSuper() || isInHandle())
      {
         returns = returns.toUpperCase();
      }
      else
      {
         returns = returns.toLowerCase();
      }
      
      StringBuilder sb = new StringBuilder();
      
      if (isInHandle())
      {
         type = "EXTERN";
      }
      
      sb.append(type).append(",");
      sb.append(returns);
      
      if (parameters.size() == 0)
      {
         sb.append(",");
      }
      else
      {
         int idx = 0;
         
         for (Parameter p : parameters)
         {
            idx++;
            
            sb.append(",");
            
            // we can't just use the raw mode string, we must use the version that is meant
            // for the signature string since RETURN is reported by the 4gl as OUTPUT
            String pmode = p.getSignatureMode();
            
            if (pmode.length() > 0)
            {
               sb.append(pmode).append(" ");
            }

            if ("BUFFER".equalsIgnoreCase(p.getType()))
            {
               String buffer = p.getAttribute("buffer");
               String forTable = p.getAttribute("for");

               sb.append("BUFFER").append(" ");
               sb.append(buffer).append(" ");
               sb.append(forTable);
            }
            else if ("TABLE".equalsIgnoreCase(p.getType()))
            {
               String table = p.getAttribute("table");
               String append = p.getAttribute("append");

               sb.append("TABLE").append(" ");
               sb.append(table);
               
               if ("true".equalsIgnoreCase(append))
               {
                  sb.append(" ").append("APPEND");
               }
            }
            else
            {
               String pname = p.getLegacyName();
               
               if (pname == null || pname.length() == 0)
               {
                  pname = String.format("arg%d", idx);
               }

               sb.append(pname).append(" ");
               sb.append(p.getType());
               
               // emit extent info
               if (p.isExtent())
               {
                  sb.append(" EXTENT");
                  if (p.isFixedExtent())
                  {
                     sb.append(" ").append(p.getExtent());
                  }
               }
            }
         }
      }

      return sb.toString();
   }
}
